#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fileio.h"
#include "kohn_gif.h"

/**************************************************
 * Project: libkohn_gif
 *  Author: Michael A. Kohn
 *   Email: mike@mikekohn.net
 *     Web: http://www.mikekohn.net/
 * License: Copyright 2004-2009 under LGPL
 **************************************************/

unsigned char data_stream[256];
int last_code;
int max_colors;
int data_size;
unsigned char color;
int first_time;

struct gif_info_t *kgif_alloc_compress()
{
struct gif_info_t *gif_info;

  gif_info=(struct gif_info_t *)malloc(sizeof(struct gif_info_t));
  memset(gif_info,0,sizeof(struct gif_info_t));

  gif_info->cnode=malloc(sizeof(struct compress_node_t)*4097);
  gif_info->transparency_index=-1;

  return gif_info;
}

void kgif_compress_reset(struct gif_info_t *gif_info)
{
  memset(gif_info,0,sizeof(struct gif_info_t));
  gif_info->transparency_index=-1;
}

int kgif_write_header(struct gif_info_t *gif_info)
{
int c,count;

/*
  for (c=0; c<colors; c++)
  {
    gif_info->color_map[c]=palette[c];
  }

  gif_info->colors=colors;
*/

  count=gif_info->colors;
  c=1;
  while(c<8)
  {
    if ((count>>c)==0) break;
    c++;
  }

  if (c<2) c=2; /* check this later */

  max_colors=1<<c;
  gif_info->code_size=c;

#ifdef DEBUG
printf("colors=%d  bits=%d\n",gif_info->colors,c);
#endif

  /* WRITE GIF HEADER */

  if (gif_info->transparency_index==-1)
  { write_chars(gif_info->fp,"GIF87a"); }
    else
  { write_chars(gif_info->fp,"GIF89a"); }
  write_int16(gif_info->fp,gif_info->width);
  write_int16(gif_info->fp,gif_info->height);
  /* putc(128+((max_colors-1)<<4)+(max_colors-1),gif_info->fp); */
  putc(128+((gif_info->code_size-1)<<4)+(gif_info->code_size-1),gif_info->fp);
  putc(0,gif_info->fp);
  putc(0,gif_info->fp);

  for (c=0; c<max_colors; c++)
  {
    putc((gif_info->color_map[c]>>16),gif_info->fp);
    putc(((gif_info->color_map[c]>>8)&255),gif_info->fp);
    putc((gif_info->color_map[c]&255),gif_info->fp);
  } 

  if (gif_info->transparency_index!=-1)
  {
    putc('!',gif_info->fp);
    putc(0xf9,gif_info->fp);
    putc(4,gif_info->fp);
    if (gif_info->transparency_index!=-1)
    { putc(1,gif_info->fp); }
      else
    { putc(0,gif_info->fp); }
    putc(0,gif_info->fp);
    putc(0,gif_info->fp);
    putc(gif_info->transparency_index&255,gif_info->fp);
    putc(0,gif_info->fp);
  }

  putc(',',gif_info->fp);
  write_int16(gif_info->fp,0);
  write_int16(gif_info->fp,0);
  write_int16(gif_info->fp,gif_info->width);
  write_int16(gif_info->fp,gif_info->height);
  putc((gif_info->code_size-1),gif_info->fp);

  putc(gif_info->code_size,gif_info->fp);

  for (c=0; c<max_colors; c++)
  {
    gif_info->cnode[c].color=c;
    gif_info->cnode[c].right=c+1;
    gif_info->cnode[c].down=-1;
  }

  /* gif_info->cnode[c-1].right=-1; */

  gif_info->clear_code=max_colors;
  gif_info->eof_code=max_colors+1;

#ifdef DEBUG
printf("clear_code=%d\n",gif_info->clear_code);
printf("eof_code=%d\n",gif_info->eof_code);
#endif

  gif_info->start_table_size=max_colors+2;
  gif_info->next_code=gif_info->start_table_size;
  gif_info->curr_code_size=gif_info->code_size+1;
  gif_info->code=-1;
  data_size=0;

#ifdef DEBUG
printf("curr_code_size=%d\n",gif_info->curr_code_size);
#endif

  gif_info->holding=gif_info->clear_code;
  gif_info->bitptr=gif_info->curr_code_size;
  first_time=1;

#ifdef DEBUG
printf("code=%d color=%d\n",gif_info->code,color);
#endif

  return 0;
}

int kgif_compress_block(struct gif_info_t *gif_info, unsigned char *uncomp_data, int len)
{
int c;
unsigned int ptr;
int t;

  ptr=0;

  if (first_time==1)
  {
    color=uncomp_data[ptr++];
    gif_info->code=color;
    last_code=gif_info->code;
    first_time=0;
  }

  while (ptr<len)
  {
    color=uncomp_data[ptr++];

#ifdef DEBUG
printf("code=%d color=%d\n",gif_info->code,color);
#endif

    if (gif_info->cnode[gif_info->code].down==-1)
    {
      last_code=gif_info->code;
      gif_info->cnode[gif_info->code].down=gif_info->next_code;
    }
      else
    {
      last_code=gif_info->code;
      gif_info->code=gif_info->cnode[gif_info->code].down;

      while(gif_info->cnode[gif_info->code].color!=color && gif_info->cnode[gif_info->code].right!=-1)
      { gif_info->code=gif_info->cnode[gif_info->code].right; }

      if (gif_info->cnode[gif_info->code].color!=color && gif_info->cnode[gif_info->code].right==-1)
      { gif_info->cnode[gif_info->code].right=gif_info->next_code; }
        else
      { continue; }
    }

#ifdef DEBUG
printf("OUTPUT CODE %d  next_code=%d  color=%d\n",last_code,gif_info->next_code,color);
#endif

    gif_info->holding=gif_info->holding+(last_code<<gif_info->bitptr);
    gif_info->bitptr=gif_info->bitptr+gif_info->curr_code_size;

    gif_info->cnode[gif_info->next_code].right=-1;
    gif_info->cnode[gif_info->next_code].down=-1;
    gif_info->cnode[gif_info->next_code].color=color;
    gif_info->code=color;


    if ((gif_info->next_code>>gif_info->curr_code_size)!=0)
    {
      if (gif_info->curr_code_size>=12)
      {
        gif_info->holding=gif_info->holding+(gif_info->clear_code<<gif_info->bitptr);
        gif_info->bitptr=gif_info->bitptr+gif_info->curr_code_size;

        for (c=0; c<max_colors; c++)
        { gif_info->cnode[c].down=-1; }

        gif_info->next_code=gif_info->start_table_size-1;
        gif_info->curr_code_size=gif_info->code_size;
        /* gif_info->code=0; */
      }
      gif_info->curr_code_size++;
    }

    gif_info->next_code++;

#ifdef DEBUG
printf("holding=0x%x  bitptr=%d  code_size=%d\n",gif_info->holding,gif_info->bitptr,gif_info->curr_code_size);
#endif

    while (gif_info->bitptr>=8)
    {
      data_stream[data_size++]=gif_info->holding&255;
      gif_info->holding=gif_info->holding>>8;
      gif_info->bitptr=gif_info->bitptr-8;
      if (data_size>=255)
      {
        putc(data_size,gif_info->fp);
        t=0;
        while(t!=data_size)
        { t=t+fwrite(data_stream,1,data_size-t,gif_info->fp); }
        data_size=0;
      }
    }
  }

  return 0;
}

int kgif_write_footer(struct gif_info_t *gif_info)
{
int t;

#ifdef DEBUG
printf("Finishing up %d\n",gif_info->code);
#endif

  gif_info->holding=gif_info->holding+(gif_info->code<<gif_info->bitptr);
  gif_info->bitptr=gif_info->bitptr+gif_info->curr_code_size;
  gif_info->holding=gif_info->holding+(gif_info->eof_code<<gif_info->bitptr);
  gif_info->bitptr=gif_info->bitptr+gif_info->curr_code_size;

  while (gif_info->bitptr>0)
  {
    data_stream[data_size++]=gif_info->holding&255;
    gif_info->holding=gif_info->holding>>8;
    gif_info->bitptr=gif_info->bitptr-8;
    if (data_size>=255)
    {
      putc(data_size,gif_info->fp);
      t=0;
      while(t!=data_size)
      { t=t+fwrite(data_stream,1,data_size-t,gif_info->fp); }
      data_size=0;
    }
  }

#ifdef DEBUG
printf("data_size=%d bitptr=%d\n",data_size,gif_info->bitptr);
#endif
  if (data_size>0)
  {
    putc(data_size,gif_info->fp);
    t=0;
    while(t!=data_size)
    { t=t+fwrite(data_stream,1,data_size-t,gif_info->fp); }
  }

  putc(0,gif_info->fp);
  putc(';',gif_info->fp);

  return 0;
}

static const char *version="2009-01-27";

const char *kgif_version_string()
{
  return version;
}


