#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kohn_gif.h"
#include "fileio.h"

/**************************************************
 * Project: libkohn_gif
 *  Author: Michael A. Kohn
 *   Email: mike@mikekohn.net
 *     Web: http://www.mikekohn.net/
 * License: Copyright 2004-2009 under LGPL
 **************************************************/

int main(int argc, char *argv[])
{
/* FILE *out,*in; */
struct gif_info_t *input_gif;
struct gif_info_t *output_gif;
unsigned char *output_line;
int x0,y0,x1,y1,x,y;
int num_bytes,ptr;
int t;

  printf("\nGIF CROP\n");
  printf("Copyright 2004-2007 - Michael Kohn <mike@mikekohn.net>\n");
  printf("Version Date: February 25, 2007\n\n");

  if (argc<7)
  {
    printf("Usage: gif_crop <input file> <output file> <x0> <y0> <x1> <y1>\n");
    exit(0);
  }

/*
  memset(&input_gif,0,sizeof(struct gif_info_t));
  memset(&output_gif,0,sizeof(struct gif_info_t));
*/

  input_gif=kgif_alloc_decompress();

  x0=atoi(argv[3]);
  y0=atoi(argv[4]);
  x1=atoi(argv[5]);
  y1=atoi(argv[6]);

  input_gif->fp=fopen(argv[1],"rb");
  if (input_gif->fp==0)
  {
    printf("Couldn't open input file.\n");
    exit(1);
  }

  if (kgif_read_header(input_gif)!=0)
  {
    printf("Input file is not a valid GIF.\n");
    fclose(input_gif->fp);
    exit(1);
  }

  if (x0<0 || x0>x1 || y0<0 || y0>y1 ||
       x1>=input_gif->width || y1>=input_gif->height)
  {
    printf("x0, y0, x1, or y1 values are out of range for this image or just bad.\n");
    fclose(input_gif->fp);
    exit(1);
  }

  output_gif=kgif_alloc_compress();
  output_gif->fp=fopen(argv[2],"wb");
  if (output_gif->fp==0)
  {
    printf("Couldn't open output file.\n");
    fclose(input_gif->fp);
    exit(1);
  }

  printf("Opening a GIF of size %d x %d\n",input_gif->width,input_gif->height);

  output_gif->width=x1-x0+1;
  output_gif->height=y1-y0+1;
  output_gif->colors=input_gif->colors;
  output_gif->transparency_index=input_gif->transparency_index;

  printf("Creating a GIF of size %d x %d\n",output_gif->width,output_gif->height);

  for (x=0; x<input_gif->colors; x++)
  { output_gif->color_map[x]=input_gif->color_map[x]; }

  kgif_write_header(output_gif);

  output_line=(unsigned char *)malloc(input_gif->width);

  /* Now for the real decompression and compression stuff */

  ptr=0;
  x=0;
  y=0;

  while(y<input_gif->height)
  {
    if (input_gif->y>y)
    { break; }

    num_bytes=kgif_decompress(input_gif);

    for (t=0; t<num_bytes; t++)
    {
      if (x>=x0 && x<=x1 && y>=y0 && y<=y1)
      {
        output_line[ptr++]=input_gif->decomp_buff[t];
        if (ptr>=output_gif->width)
        {
          kgif_compress_block(output_gif,output_line,output_gif->width);
          ptr=0;
        }
      }

      x++;
      if (x>=input_gif->width)
      {
        x=0;
        y++;
      }
    }
  }

  kgif_write_footer(output_gif);
  free(output_line);

  fclose(output_gif->fp);
  fclose(input_gif->fp);

/*
  free(output_gif->cnode);
  output_gif->cnode=0;
*/

  kgif_info_free(input_gif);
  kgif_info_free(output_gif);

  return 0;
}



