#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "fileio.h"
#include "kohn_gif.h"

/**************************************************
 * Project: libkohn_gif
 *  Author: Michael A. Kohn
 *   Email: mike@mikekohn.net
 *     Web: http://www.mikekohn.net/
 * License: Copyright 2004-2009 under LGPL
 **************************************************/

/* #define DECOMP_BUFFER_SIZE 32768 */

struct gif_info_t *kgif_alloc_decompress()
{
struct gif_info_t *gif_info;

  gif_info=(struct gif_info_t *)malloc(sizeof(struct gif_info_t));
  memset(gif_info,0,sizeof(struct gif_info_t));

  /* 4096 should be enough, but just incase do 4097 */
  gif_info->dnode=(struct decompress_node_t *)malloc(sizeof(struct decompress_node_t)*4097);
  gif_info->decomp_buff=(unsigned char *)malloc(DECOMP_BUFFER_SIZE);

  return gif_info;
}

int kgif_info_free(struct gif_info_t *gif_info)
{
  if (gif_info->dnode!=0) free(gif_info->dnode);
  gif_info->dnode=0;

  if (gif_info->cnode!=0) free(gif_info->cnode);
  gif_info->cnode=0;

  if (gif_info->decomp_buff!=0) free(gif_info->decomp_buff);
  gif_info->decomp_buff=0;

  free(gif_info);

  return 0;
}

int get_color(char *s)
{
int t,l,n;

  l=strlen(s);
  n=0;

  for (t=0; t<l; t++)
  {
    s[t]=tolower(s[t]);

    if (s[t]>='0' && s[t]<='9')
    {
      n=(n<<4)+(s[t]-'0');
    }
      else
    if (s[t]>='a' && s[t]<='f')
    {
      n=(n<<4)+(s[t]-'a'+10);
    }
  }

  return n;
}

int kgif_read_header(struct gif_info_t *gif_info)
{
FILE *in;
int ch,t,x,y,r,g,b;
char temp_buffer[7];

  gif_info->transparency_index=-1;

  in=gif_info->fp;

  read_chars(in,temp_buffer,6);

  if (strcmp(temp_buffer,"GIF89a")==0)
  { gif_info->version=89; }
    else
  if (strcmp(temp_buffer,"GIF87a")==0)
  { gif_info->version=87; }
    else
  {
    printf("Image is not a GIF.\n");
    return -1;
  }

  gif_info->width=read_int16(in);
  gif_info->height=read_int16(in);
  gif_info->resolution=getc(in);
  gif_info->bgcolor=getc(in);
  gif_info->aspect=getc(in);
  gif_info->colors=1<<((gif_info->resolution&7)+1);

#ifdef DEBUG
printf("GIF HEADER\n");
printf("-----------------------------------------\n");
printf("  Signature: %s\n",temp_buffer);
printf(" Image size: %d x %d\n",gif_info->width,gif_info->height);
printf(" Resolution: 0x%02x colors=%d\n",gif_info->resolution,gif_info->colors);
printf("    BgColor: 0x%02x\n",gif_info->bgcolor);
printf("     Aspect: 0x%02x\n\n",gif_info->aspect);
#endif

  /* IF THERE IS A GLOBAL COLOR MAP, READ IT */

  if ((gif_info->resolution&128)==128)
  {
    for (t=0; t<gif_info->colors; t++)
    {
      r=getc(in);
      g=getc(in);
      b=getc(in);
      gif_info->color_map[t]=(r<<16)+(g<<8)+b;
      gif_info->bw_color_map[t]=(unsigned char)((int)(r+g+b)/3);
    }
  }

  /* READ EXTENSION BLOCKS */

  while(1)
  {
    ch=getc(in);
    if (ch=='!')
    {
      ch=getc(in);

      if (ch==0xf9)
      {
        getc(in);    /* Size, should always be 4 */
        t=getc(in);  /* Packed fields */
        getc(in);    /* Delay time (2 bytes) */
        getc(in);
        if ((t&1)==1)
        { gif_info->transparency_index=getc(in); }
          else
        { getc(in); }
#ifdef DEBUG
        printf("Transparency Index is %d\n",gif_info->transparency_index);
#endif
        getc(in);
      }
        else
      {
        while(1)
        {
          ch=getc(in);
          if (ch==0) break;
          for (t=0; t<ch; t++) getc(in);
        }
      }
    }
      else
    { ungetc(ch,in); break; }
  }

  /* READ IN IMAGE DESCRIPTOR BLOCK */

  t=getc(in);

  if (t!=',')
  {
    printf("Expected ',' got %c at 0x%06x in image 1\n",t,(int)ftell(in));
    return -1;
  }

  x=read_int16(in);
  y=read_int16(in);
  gif_info->width=read_int16(in);
  gif_info->height=read_int16(in);
  gif_info->flags=getc(in);

#ifdef DEBUG
printf("GIF IMAGE DESCRIPTION\n");
printf("-----------------------------------------\n");
printf("Coordinates: (%d,%d)\n",x,y);
printf(" Image size: %d x %d\n",gif_info->width,gif_info->height);
printf("      Flags: 0x%02x\n\n",gif_info->flags);
#endif

  /* IF THERE IS A LOCAL COLOR MAP, READ IT */

  if ((gif_info->flags&128)==128)
  {
    /* gif_info.colors=1<<((gif_info.flags&7)+1); */
    x=1<<((gif_info->flags&7)+1);

    for (t=0; t<x; t++)
    {
      r=getc(in);
      g=getc(in);
      b=getc(in);
      gif_info->color_map[t]=(r<<16)+(g<<8)+b;
      gif_info->bw_color_map[t]=(unsigned char)((int)(r+g+b)/3);
    }
  }

  gif_info->code_size=getc(in);
  gif_info->start_table_size=(1<<gif_info->code_size);

#ifdef DEBUG
printf("code_size=%d start_table_size=%d\n",gif_info->code_size,gif_info->start_table_size);
#endif

  for (t=0; t<gif_info->start_table_size; t++)
  {
    gif_info->dnode[t].prev=-1;
    gif_info->dnode[t].color=t; 
  }

  gif_info->clear_code=t;
  gif_info->eof_code=t+1;
  gif_info->next_code=t+2;
  gif_info->code_size++;

  gif_info->x=0; gif_info->y=0;
  gif_info->holding=0;
  gif_info->bitptr=0;
  gif_info->curr_code_size=gif_info->code_size;
  gif_info->old_code=-1;

  gif_info->mask=(1<<gif_info->curr_code_size)-1;
  gif_info->data_count=0;

  gif_info->full_image_size=gif_info->width*gif_info->height;

  /* gif_info->accum_ptr=0; */
  gif_info->ptr=0;
  gif_info->status=0;

  return 0;
}

int kgif_decompress(struct gif_info_t *gif_info)
{
int t,r;
int total_bytes;
unsigned char temp_buffer[4097];

  total_bytes=0;

  /* while(1) */
  {
    /* Decompress Image */

    if (gif_info->status==0)
    {
      while(gif_info->bitptr<gif_info->curr_code_size)
      {
        if (gif_info->data_count==0)
        {
          if ((gif_info->data_count=getc(gif_info->fp))==EOF) break;
          if (gif_info->data_count==0) break;
        }
        t=getc(gif_info->fp);
        gif_info->holding=gif_info->holding+(t<<gif_info->bitptr);
        gif_info->data_count--;
        gif_info->bitptr=gif_info->bitptr+8;
      }

/*
if (t==EOF || data_count1==EOF)
{
  printf("EOF! %d\n",ptr1+accum_ptr1);
  break;
}
*/

      gif_info->code=gif_info->holding&gif_info->mask;
      gif_info->holding=gif_info->holding>>gif_info->curr_code_size;
      gif_info->bitptr=gif_info->bitptr-gif_info->curr_code_size;

      if (gif_info->code==gif_info->clear_code)
      {
        gif_info->curr_code_size=gif_info->code_size;
        gif_info->next_code=gif_info->start_table_size+2;
        gif_info->old_code=-1;
        gif_info->mask=(1<<gif_info->curr_code_size)-1;
        return 0;
        /* continue; */
      }
        else
      if (gif_info->old_code==-1)
      {
        /* gif_info->decomp_buff[gif_info->ptr++]=gif_info->dnode[gif_info->code].color; */
        gif_info->decomp_buff[total_bytes++]=gif_info->dnode[gif_info->code].color;
      }
        else
      if (gif_info->code==gif_info->eof_code)
      {
        gif_info->status=1;
      }
        else
      {
        if (gif_info->code<gif_info->next_code)
        {
          t=0;
          r=gif_info->code;
          while(1)
          {
            temp_buffer[t++]=gif_info->dnode[r].color;
            if (gif_info->dnode[r].prev==-1) break;
            r=gif_info->dnode[r].prev;
          }
  
          for (r=t-1; r>=0; r--)
          {
            /* gif_info->decomp_buff[gif_info->ptr++]=gif_info->temp_buffer[r]; */
            gif_info->decomp_buff[total_bytes++]=temp_buffer[r];
          }
  
          gif_info->dnode[gif_info->next_code].color=temp_buffer[t-1];
          gif_info->dnode[gif_info->next_code].prev=gif_info->old_code;
  
          if (gif_info->next_code==gif_info->mask && gif_info->mask!=4095)
          {
            gif_info->curr_code_size++;
            gif_info->mask=(1<<gif_info->curr_code_size)-1;
          }
          gif_info->next_code++;
        }
          else
        {
          t=0;
          r=gif_info->old_code;
          while(1)
          {
            temp_buffer[t++]=gif_info->dnode[r].color;
            if (gif_info->dnode[r].prev==-1) break;
            r=gif_info->dnode[r].prev;
          }

          gif_info->dnode[gif_info->next_code].color=temp_buffer[t-1];
          gif_info->dnode[gif_info->next_code].prev=gif_info->old_code;

          if (gif_info->next_code==gif_info->mask && gif_info->mask!=4095)
          {
            gif_info->curr_code_size++;
            gif_info->mask=(1<<gif_info->curr_code_size)-1;
          }

          gif_info->next_code++;

          for (r=t-1; r>=0; r--)
          {
            /* gif_info->decomp_buff[gif_info->ptr++]=gif_info->temp_buffer[r]; */
            gif_info->decomp_buff[total_bytes++]=temp_buffer[r];
          }

          /* gif_info->decomp_buff[gif_info->ptr++]=gif_info->temp_buffer[t-1]; */
          gif_info->decomp_buff[total_bytes++]=temp_buffer[t-1];
        }
      }

      gif_info->old_code=gif_info->code;
    }
      else
    if (gif_info->status==1)
    {
      /* gif_info->decomp_buff[gif_info->ptr++]=gif_info->bgcolor; */
      gif_info->decomp_buff[total_bytes++]=gif_info->bgcolor;
    }

    /* if (gif_info->accum_ptr>gif_info->full_image_size) gif_info->status=3; */

  }

  gif_info->count=total_bytes;
  gif_info->ptr=0;

  return total_bytes;
}

